/*
 *  server_example2.c
 *
 *  Copyright 2013 Michael Zillgith
 *
 *  This file is part of libIEC61850.
 *
 *  libIEC61850 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libIEC61850 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  See COPYING file for the complete license text.
 */

#include "server_api_v2.h"
#include "thread.h"
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

/* Include the generated header with the model access handles */
#include "static_model.h"

/* import IEC 61850 device model created from SCL-File */
extern IedModel iedModel;

static int running = 0;

void sigint_handler(int signalId)
{
	running = 0;
}

int main(int argc, char** argv) {

	IedServer iedServer = IedServer_create(&iedModel);

	// get stored values from persistent storage

	// set initial measurement and status values from process

	/* MMS server will be instructed to start listening to client connections. */
	IedServer_start(iedServer, 102);

	if (!IedServer_isRunning(iedServer)) {
		printf("Starting server failed! Exit.\n");
		IedServer_destroy(iedServer);
		exit(-1);
	}

	running = 1;

	signal(SIGINT, sigint_handler);

	float power = 500.f;

	MmsValue* powerValue = MmsValue_newFloat(power);
	MmsValue* powerTimestamp = MmsValue_newUtcTime(time(NULL));


	while (running) {
		MmsValue_setUtcTime(powerTimestamp, time(NULL));

		IedServer_lockDataModel(iedServer);

		IedServer_updateAttributeValue(iedServer, IEDMODEL_Inverter_MMXU1_TotW_mag_f, powerValue);
		IedServer_updateAttributeValue(iedServer, IEDMODEL_Inverter_MMXU1_TotW_t, powerTimestamp);

		IedServer_unlockDataModel(iedServer);

		power += 0.1f;

		MmsValue_setFloat(powerValue, power);

		Thread_sleep(500);
	}

	MmsValue_delete(powerValue);
	MmsValue_delete(powerTimestamp);

	/* stop MMS server - close TCP server socket and all client sockets */
	IedServer_stop(iedServer);

	/* Cleanup - free all resources */
	IedServer_destroy(iedServer);
} /* main() */
