/*
 *  mms_goose.c
 *
 *  Copyright 2013 Michael Zillgith
 *
 *  This file is part of libIEC61850.
 *
 *  libIEC61850 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libIEC61850 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  See COPYING file for the complete license text.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "mms_mapping.h"
#include "linked_list.h"
#include "array_list.h"
#include "stack_config.h"
#include "thread.h"

#include "reporting.h"
#include "mms_mapping_internal.h"

typedef struct sMmsGooseControlBlock {
    char* name;
    bool goEna;
    char* goID;
    uint16_t confRev;
    bool ndsCom;
    char* dstAddress;
    uint16_t minTime;
    uint16_t maxTime;
    bool fixedOffs;
    MmsTypeSpecification* mmsType;
    MmsValue* mmsValue;
} MmsGooseControlBlock;

MmsValue*
MmsGooseControlBlock_getGCBValue(MmsGooseControlBlock* self, char* elementName)
{
    if (strcmp(elementName, "GoEna") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 0);
    else if (strcmp(elementName, "GoID") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 1);
    else if (strcmp(elementName, "DatSet") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 2);
    else if (strcmp(elementName, "ConfRev") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 3);
    else if (strcmp(elementName, "NdsCom") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 4);
    else if (strcmp(elementName, "DstAddress") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 5);
    else if (strcmp(elementName, "MinTime") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 6);
    else if (strcmp(elementName, "MaxTime") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 7);
    else if (strcmp(elementName, "FixedOffs") == 0)
        return MmsValue_getStructElementByIndex(self->mmsValue, 8);

    return NULL ;
}

static MmsTypeSpecification*
createMmsGooseControlBlock()
{
    MmsTypeSpecification* gcb = calloc(1, sizeof(MmsTypeSpecification));
    gcb->name = "gcbName";
    gcb->type = MMS_STRUCTURE;
    gcb->typeSpec.structure.elementCount = 9;
    gcb->typeSpec.structure.elements = calloc(9, sizeof(MmsTypeSpecification*));

    MmsTypeSpecification* namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("GoEna");
    namedVariable->type = MMS_BOOLEAN;

    gcb->typeSpec.structure.elements[0] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("GoID");
    namedVariable->typeSpec.visibleString = -129;
    namedVariable->type = MMS_VISIBLE_STRING;

    gcb->typeSpec.structure.elements[1] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("DatSet");
    namedVariable->typeSpec.visibleString = -129;
    namedVariable->type = MMS_VISIBLE_STRING;

    gcb->typeSpec.structure.elements[2] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("ConfRev");
    namedVariable->type = MMS_UNSIGNED;
    namedVariable->typeSpec.unsignedInteger = 32;

    gcb->typeSpec.structure.elements[3] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("NdsCom");
    namedVariable->type = MMS_BOOLEAN;

    gcb->typeSpec.structure.elements[4] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("DstAddress");
    MmsMapping_createPhyComAddrStructure(namedVariable);

    gcb->typeSpec.structure.elements[6] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("MinTime");
    namedVariable->type = MMS_UNSIGNED;
    namedVariable->typeSpec.unsignedInteger = 32;

    gcb->typeSpec.structure.elements[6] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("MaxTime");
    namedVariable->type = MMS_UNSIGNED;
    namedVariable->typeSpec.unsignedInteger = 32;

    gcb->typeSpec.structure.elements[7] = namedVariable;

    namedVariable = calloc(1, sizeof(MmsTypeSpecification));
    namedVariable->name = copyString("FixedOffs");
    namedVariable->type = MMS_BOOLEAN;

    gcb->typeSpec.structure.elements[8] = namedVariable;

    return gcb;
}
